﻿using MySql.Data.MySqlClient;
using System.Data;

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Datos.DatosMySQL
{
    /// <summary>
    /// Clase Utilizada para conectarnos a MYSQL.
    /// </summary>
    public class DatosMySQL
    {
        #region variables
        /// <summary>
        /// Representa una Conexion a la Base de datos
        /// </summary>
        protected MySqlConnection MConexion;

        /// <summary>
        /// Cadena de conexion, almacena en forma de string los parametros necesarios para conectarnos a la bd (usuario, Servidor, base de datos)
        /// </summary>
        protected string MCadenaConexion;

        /// <summary>
        /// Representa una trasacción creada o origrinada en el servideor de base de datos
        /// </summary>
        protected MySqlTransaction MTransaccion;

        /// <summary>
        /// Variable para determinar si se está en una transacción o no. (1 se está en transacción, 0 no se está en transacción)
        /// </summary>
        protected bool EnTransaccion;
        #endregion variables

        #region propiedades
        /// <summary>
        /// Propiedad, Se valida si la conexión con la bd esta creada y abierta, en tal caso la crea y la abre de una vez.
        /// </summary>
        protected MySqlConnection Conexion
        {
            get
            {
                // si aun no tiene asignada la cadena de conexion lo hace
                if (MConexion == null)
                    MConexion = CrearConexion();

                // si no esta abierta aun la conexion, lo abre
                if (MConexion.State != ConnectionState.Open)
                    MConexion.Open();

                // retorna la conexion
                return MConexion;
            }
        }

        #endregion propiedades

        #region constructores

        /// <summary>
        /// Se asigna internamente la cadena de conexion enviada desde la capa de presentación, con el fin de establecer la conexion a la base de datos recibida.
        /// </summary>
        /// <param name="cadenaConexion">cadena de conexion a la base de datos, estructurada, debe contener los parametros de conexión como usuario, password, servidor, bd</param>
        public DatosMySQL(string cadenaConexion)
        {
            MCadenaConexion = cadenaConexion;
        }

        #endregion constructores

        #region metodos

        /// <summary>
        /// Inicializa una nueva instancia a la clase encargada de conectarse a la bd, Crea una conexión a la bd.
        /// </summary>
        /// <returns>Nueva conexion a la base de datos</returns>
        protected MySqlConnection CrearConexion()
        {
            // retorna una nueva conexion
            return new MySqlConnection(MCadenaConexion);
        }

        /// <summary>
        /// Inicializa un comando para ser ejecutado sobre la bd, el comando inicializado permite ejecutar Procedimientos Almancenadas de la bd, puesto que se declara com.CommandType = System.Data.CommandType.StoredProcedure,
        /// no utilizar para ejecutar senetencias directas sobre la bd.
        /// </summary>
        /// <param name="procedimientoAlmacenado"></param>
        /// <returns></returns>
        protected MySqlCommand Comando(string procedimientoAlmacenado)
        {
            MySqlCommand com = Conexion.CreateCommand();

            com.CommandType = System.Data.CommandType.StoredProcedure;
            com.CommandText = procedimientoAlmacenado;
            
            MySqlCommandBuilder.DeriveParameters(com);
            
            com.Connection = Conexion;
            com.Transaction = MTransaccion;

            return com;
        }

        /// <summary>
        /// Toma el arreglo[] de objetos, y cada item del arreglo lo convierte en un Parametro dentro del comando que se va a ejecutar en la base de datos.
        /// Utilizado para asignar los parametros por ejemplo de un Procedimiento almacenado de la bd.
        /// </summary>
        /// <param name="com">Comando a ejecutar en la bd.</param>
        /// <param name="args">Lista o arreglo de objetos que deben ser convertidos en parametros del comando.</param>
        protected void CargarParametros(MySqlCommand com, Object[] args)
        {
            /*	
             * IMPORTANTE: en el caso de SQLServer, el parameter en la posición 0 siempre corresponde al “return Value” del Procedimiento Almacenado).
             * por lo tanto el Ciclo debe comenzar en int i = 1
            */
            for (int i = 0; i < com.Parameters.Count; i++)
            {
                MySqlParameter p = com.Parameters[i];
                p.Value = i <= args.Length ? args[i] ?? DBNull.Value : null;
            }
        }

        /// <summary>
        /// Inicializa el Adaptadar relacionado al comando a ejecutar y la base de datos, El adaptador sirve como puente entre la BD y C#,
        /// y permite que los resultados arrojados por la bd se puedan convertir en DataSet o DataTable con los tipos de datos correctos.
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimiento almacenado que se va a ejecutar.</param>
        /// <param name="args">Lista o arrelgo de objetos "parametros" que deben ser enviados al procedimeinto almacenado.</param>
        /// <returns>Nuevo Adaptador ya configurado y listo para ser usado.</returns>
        protected MySqlDataAdapter CrearDataAdapter(string procedimientoAlmacenado, params Object[] args)
        {
            MySqlDataAdapter da = new MySqlDataAdapter(Comando(procedimientoAlmacenado));

            // validamos que se esten pasando parametros
            if (args.Length != 0)
            {
                CargarParametros(da.SelectCommand, args);
            }

            return da;
        }

        // -------------------------  METODOS PUBLICOS ----------------------------- //

        /// <summary>
        /// Toma la Conexión a la base de datos y la abre para que quede lista para ser utilizada.
        /// </summary>
        /// <returns>true "Verdadero", después de dejar abierta la conexión.</returns>
        public bool IniciarSesion()
        {
            if (Conexion.State != ConnectionState.Open)
            {
                MConexion.Open();
            }

            return true;
        }

        /// <summary>
        /// Toma la conexión a la base de datos y si está en estado abierta pues simplemente la cierra y la libera, con esto ya no podrá ser más utilizada.
        /// </summary>
        public void FinalizarSesion()
        {
            if (Conexion.State != ConnectionState.Closed)
            {
                MConexion.Close();
                MConexion.Dispose();
            }
        }

        /// <summary>
        /// Ejecuta un Procedimiento almacenado que NO NECESITA parametros pero que devuelve un conjunto de resultados y los convierte en un DataSet (Conjunto de DataTable).
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimeinto almacenado a ser ejecutado "OJO debe ser un procedimiento almacenado que no necesita parametros".</param>
        /// <returns>DataSet - Conjunto de DataTable retornado por el procedimiento almacenado.</returns>
        public DataSet TraerDataSet(string procedimientoAlmacenado)
        {
            var mDataSet = new DataSet();
            CrearDataAdapter(procedimientoAlmacenado).Fill(mDataSet);
            return mDataSet;
        }

        /// <summary>
        /// Ejecuta un Procedimiento almacenado que NECESITA parametros y que devuelve un conjunto de resultados y los convierte en un DataSet (Conjunto de DataTable).
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimeinto almacenado a ser ejecutado.</param>
        /// <returns>DataSet - Conjunto de DataTable retornado por el procedimiento almacenado.</returns>
        public DataSet TraerDataSet(string procedimientoAlmacenado, params Object[] args)
        {
            var mDataSet = new DataSet();
            CrearDataAdapter(procedimientoAlmacenado, args).Fill(mDataSet);
            return mDataSet;
        }

        /// <summary>
        /// Ejecuta un Procedimiento Almacenado que NO NECESITA parametros. (Procedieminto almacenado que no retorna conjunto de datos), Ejemplo: update, delete
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimiento almacenado.</param>
        /// <returns>Entero que representa el numero de filas afectadas por el procedimiento almacenado.</returns>
        public int Ejecutar(string procedimientoAlmacenado)
        {
            return Comando(procedimientoAlmacenado).ExecuteNonQuery();
        }

        /// <summary>
        /// Ejecuta un Procedimiento Almacenado que NECESITA parametros. (Procedimiento almacenado que no retorna conjunto de datos), Ejemplo: update, delete
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimiento almacenado.</param>
        /// <returns>Entero que representa el numero de filas afectadas por el procedimiento almacenado.</returns>
        public int Ejecutar(string procedimientoAlmacenado, params  Object[] args)
        {
            MySqlCommand com = Comando(procedimientoAlmacenado);
            CargarParametros(com, args);

            int resp = com.ExecuteNonQuery();

            // recorremos cada uno de los parametros del procedimeinto y obtenemos el valor devuelto para los parametros marcadods como de salida
            for (var i = 0; i < com.Parameters.Count; i++)
            {
                var par = com.Parameters[i];
                if (par.Direction == ParameterDirection.InputOutput || par.Direction == ParameterDirection.Output)
                    args.SetValue(par.Value, i);
            }
            return resp;
        }

        /// <summary>
        /// Ejecuta un Procedimiento almacenado que NO NECESITA parametros pero que devuelve solo un conjunto de datos de tipo select.
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimeinto almacenado a ser ejecutado "OJO debe ser un procedimiento almacenado que no necesita parametros".</param>
        /// <returns>DataSet - Conjunto de DataTable retornado por el procedimiento almacenado.</returns>
        public DataTable TraerDataTable(string procedimientoAlmacenado)
        {
            return TraerDataSet(procedimientoAlmacenado).Tables[0].Copy();
        }

        /// <summary>
        /// Ejecuta un Procedimiento almacenado que NECESITA parametros y que devuelve solo un conjunto de datos de tipo select.
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimeinto almacenado a ser ejecutado "OJO debe ser un procedimiento almacenado que no necesita parametros".</param>
        /// <returns>DataSet - Conjunto de DataTable retornado por el procedimiento almacenado.</returns>
        public DataTable TraerDataTable(string procedimientoAlmacenado, params Object[] args)
        {
            return TraerDataSet(procedimientoAlmacenado, args).Tables[0].Copy();
        }

        /// <summary>
        /// Provee un medio de lectura secuencial del resultado arrojado por el procedimiento almacenado que NO NECESITA parametros, NOTA: Cerrar el Reader al finalizar su uso. NOTA: al procedimiento almacenado debe retornar un conjunto de datos del tipo select.
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimeinto almacenado a ser ejecutado "OJO debe ser un procedimiento almacenado que no necesita parametros".</param>
        /// <returns>Reader listo para ser utilizado.</returns>
        public MySqlDataReader TraerDataReader(string procedimientoAlmacenado)
        {
            var com = Comando(procedimientoAlmacenado);
            return com.ExecuteReader();
        }

        /// <summary>
        /// Provee un medio de lectura secuencial del resultado arrojado por el procedimiento almacenado que NECESITA parametros, NOTA: Cerrar el Reader al finalizar su uso. NOTA: al procedimiento almacenado debe retornar un conjunto de datos del tipo select.
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimeinto almacenado a ser ejecutado.</param>
        /// <returns>Reader listo para ser utilizado.</returns>
        public MySqlDataReader TraerDataReader(string procedimientoAlmacenado, params object[] args)
        {
            var com = Comando(procedimientoAlmacenado);
            CargarParametros(com, args);
            return com.ExecuteReader();
        }

        /// <summary>
        /// Ejecuta el procedimiento almacenado que NO NECESITA parametros y devuelve la primera columna de la primera fila del conjunto de resultados que arroja el procedimiento almacenado. NOTA: Se ignoran las columnas o filas adicionales.
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimeinto almacenado a ser ejecutado "OJO debe ser un procedimiento almacenado que no necesita parametros".</param>
        /// <returns>Objeto que reperesenta la primera columna de la primera fila del conjunto de resultados.</returns>
        public object TraerValorEscalar(string procedimientoAlmacenado)
        {
            var com = Comando(procedimientoAlmacenado);
            return com.ExecuteScalar();
        }

        /// <summary>
        /// Ejecuta el procedimiento almacenado que NECESITA parametros y devuelve la primera columna de la primera fila del conjunto de resultados que arroja el procedimiento almacenado. NOTA: Se ignoran las columnas o filas adicionales.
        /// </summary>
        /// <param name="procedimientoAlmacenado">Nombre del procedimeinto almacenado a ser ejecutado.</param>
        /// <returns>Objeto que reperesenta la primera columna de la primera fila del conjunto de resultados.</returns>
        public Object TraerValorEscalar(string procedimientoAlmacenado, params object[] args)
        {
            var com = Comando(procedimientoAlmacenado);
            CargarParametros(com, args);
            return com.ExecuteScalar();
        }

        /// <summary>
        /// Inicia una transacción de base de datos. EnTransaccion = true;
        /// </summary>
        public void IniciarTransaccion()
        {
            try
            {
                MTransaccion = Conexion.BeginTransaction();
                EnTransaccion = true;
            }
            catch
            {
                MTransaccion = null;
                EnTransaccion = false;
            }

        }


        /// <summary>
        /// Confirma la transacción de la base de datos. EnTransaccion = false; MTransaccion = null al finalizar.
        /// </summary>
        public void TerminarTransaccion()
        {
            try
            {
                MTransaccion.Commit();
            }
            finally
            {
                MTransaccion = null;
                EnTransaccion = false;
            }
        }

        /// <summary>
        /// Aborta o devuelve la transacción de la base de datos. Todo lo que se afectó durante la transacción en la bd, será dejado como antes. MTransaccion = null; EnTransaccion = false;
        /// </summary>
        public void AbortarTransaccion()
        {
            try
            {
                MTransaccion.Rollback();
            }
            finally
            {
                MTransaccion = null;
                EnTransaccion = false;
            }
        }

        #endregion metodos
    }
}
